/* ITKA203 Käyttöjärjestelmät kevät 2014
 * Malliharjoitustyö
 *
 * Tässä on C-kielinen ohjelma, joka (jalavan x86-64 -tietokoneessa
 * gcc:llä käännettynä) tekee seuraavaa:
 *
 *      - kääntää 64-bittisiä kokonaislukuja sisältävän taulukon
 *        alkiot käänteiseen järjestykseen, esimerkiksi taulukosta
 *        {45, -4, 56, 57, -4, 9} pitäisi tulla {9, -4, 57, 56, -4,
 *        45}. Operaatio tehdään "paikallaan" eli vaihdetaan keskenään
 *        ensimmäinen ja viimeinen, toinen ja toiseksi viimeinen, ja
 *        niin edelleen.
 *
 *      - ilmoittaa paluuarvona, montako kertaa vaihdettiin sama luku
 *        itsensä kanssa; koska em. esimerkissä -4 oli sekä toinen
 *        että toiseksi viimeinen alkio, pitäisi paluuarvoksi tulla
 *        ykkönen.
 *
 * Ohjelmasta ei varmaan ole hyötyä muuhun kuin sen pedagogiseen
 * tarkoitukseen: Voimme debuggeria käyttäen nähdä, kuinka
 * aliohjelmien kutsuminen, parametrien välitys, paikallisten
 * muuttujien käyttö sekä paluuarvon palauttaminen toimii konekielen
 * tasolla.
 *
 * Kaikki muuttujat ovat tyypiltään ``long long int`` jotta ne jalavan
 * gcc:llä tulisivat kokonaisen 64-bittisen rekisterin kokoisiksi.
 * Syy myös tähän on pedagoginen (helpompi tulkita konekielistä
 * käännöstä). Käytännössä 64-bittiset luvut lienevät moneen
 * tarpeeseen kärpäsen ampumista tykillä, mutta toisaalta esimerkiksi
 * kellonajan ilmoittaminen sekunteina keskiyöstä 1.1.1970 alkaen
 * epäonnistuu karkeasti 32-bittisellä kokonaisluvulla jo aamuyöstä
 * tiistaina 19.1.2038 (tai tällä hetkellä tehtäville, vaikkapa 25
 * vuoden laina-aikaa koskeville, aikavälilaskelmille):
 * http://en.wikipedia.org/wiki/Year_2038_problem
 */

#include<stdio.h>

/**
 * Sattuneesta syystä varsinaisesta esimerkistä tulee mielekkäämpi,
 * kun siinä tehdään välissä yksi kutsu "turhaan" aliohjelmaan, jossa
 * ei tehdä mitään.
 * 
 * (Kutsu nimittäin varmistaa, että itse tutkittavana oleva aliohjelma
 * toimii "kunnon kansalaisen" tavoin, eikä suorituskykyoptimoidun
 * "eliitin", joka tietää olevansa aina kutsupinon päällimäinen;
 * lisätietoa saa googlettamalla "amd64 red zone"; tällaiset ABI:n
 * detaljit eivät oikeastaan kuulu kurssin sisältöön, joten
 * hyväksytään nyt että tässä on tällainen turhalta vaikuttava
 * aliohjelma, joka ei tee mitään muuta kuin palaa heti takaisin
 * kutsuvaan ohjelmaan);
 */
void turha_kutsu(){
}

/** 
 * Aliohjelma, joka manipuloi taulukkoa. Parametrina on C:n taulukko,
 * eli käytännössä ensimmäisen alkion muistiosoite, sekä kokonaisluku,
 * joka tulkitaan taulukon kooksi (eli alkioiden lukumääräksi).
 */
long long int kaanna_taulukko(long long int taulu[], long long int koko){
  long long int ioikea, ivasen;
  long long int lkm_samoja = 0;
  long long int tmp;
  turha_kutsu();
  ivasen=0;
  for(ioikea=koko-1; ioikea>ivasen; ioikea--){
    if (taulu[ivasen] == taulu[ioikea]){
      lkm_samoja++;
    } else {
      /* vaihto */
      tmp = taulu[ivasen];
      taulu[ivasen] = taulu[ioikea];
      taulu[ioikea] = tmp;
    }
    ivasen++;
  }
  return lkm_samoja;
}


/** Testailua varten tein pienen tulostuskoodin. */
void tulosta_taulukko(long long int taulu[], long long int koko){
  int i;
  printf("64-bittisiä lukuja %d kpl alkaen osoitteesta %llx:\n", koko, taulu);
  for(i=0; i<koko; i++){
    printf("%lld ", taulu[i]);
  }
  printf("\n");
}


/**
 * Pääohjelma (joka C:ssä on käytännössä vain sopimuksen mukaisesti
 * nimetty aliohjelma).
 */
int main(int argc, char *argv[]){
  long long int testitaulu[] = {-1, 2, 3, -16, 2, 8};
  long long int koko = 6;
  long long int samoja;

#ifndef HILJAA
  tulosta_taulukko(testitaulu, koko);
#endif

  samoja = kaanna_taulukko(testitaulu, koko);

#ifndef HILJAA
  tulosta_taulukko(testitaulu, koko);  
  printf("Samoja lukuja peilikuvasijainneissa oli %lld kpl.\n", samoja);
#endif

  return 0;
}
